"""Bakery house."""

from __future__ import annotations

from .stuff import BakeryLogger, DefaultLogger

logger: BakeryLogger | None = DefaultLogger()

# ruff: noqa: E402
from .bakery import *
from .baking import *
from .cake import *
from .piece_of_cake import *
from .stuff import *

# ruff: noqa: F405, PLE0604
__all__ = [
    *bakery.__all__,  # type: ignore[name-defined]
    *baking.__all__,  # type: ignore[name-defined]
    *cake.__all__,  # type: ignore[name-defined]
    *piece_of_cake.__all__,  # type: ignore[name-defined]
    *stuff.__all__,  # type: ignore[name-defined]
]



"""Cake.

Book, recipes, etc.
"""

from __future__ import annotations

__all__ = ["Cake", "Pastry", "__Cake__", "hand_made"]

from contextlib import AbstractAsyncContextManager, AbstractContextManager, contextmanager
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Generic,
    Literal,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import ParamSpec, Self

from .baking import BakingMethod, bake_recipe, check_baking_method, determine_baking_method
from .piece_of_cake import PieceOfCake
from .stuff import _LOGGER as logger  # noqa: N811
from .stuff import (
    CakeRecipe,
    assert_baked,
    flatten,
    is_baked,
    is_cake,
    is_cake_or_piece,
    is_piece_of_cake,
    recipe_format,
)
from .stuff.types import UNDEFINED

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterator
    from types import TracebackType


R = TypeVar("R")


class Pastry(CakeRecipe, Generic[R]):
    """Pastry is a public cake interface with almost zero name collision.

    Your item ingredients and cooking method stored here.
    """

    __CAKE_ATTRIBUTE_ERROR_NAMES__: Final[set[str]] = {
        "func",  # partial
        "__wrapped__",  # functools special attribute
    }

    def __init__(
        self,
        _cake_recipe: R,
        *_cake_recipe_args: Any,
        _cake_baking_method: BakingMethod = BakingMethod.BAKE_AUTO,
        _cake_name: str = "",
        **_cake_recipe_kwargs: Any,
    ) -> None:
        self.__cake_recipe: Final = _cake_recipe
        self.__cake_recipe_args: Final = _cake_recipe_args
        self.__cake_recipe_kwargs: Final = _cake_recipe_kwargs

        self.__cake_baking_method: BakingMethod = _cake_baking_method
        self.__cake_result: Any = None
        self.__cake_is_baked: bool = False
        self.__cake_name: str = _cake_name

        self.__cake_replaced: Pastry | None = None

        if not self.__cake_baking_method:
            self.__cake_baking_method = determine_baking_method(self.__cake_recipe)
        else:
            check_baking_method(self.__cake_recipe, _cake_baking_method)

        if self.__cake_baking_method == BakingMethod.BAKE_NO_BAKE:
            self.__cake_result = self.__cake_recipe
            self.__cake_is_baked = True

    def __set_name__(self, _: Any, name: str) -> None:
        self.__cake_name = name

    @property
    def __cake_name__(self) -> str:
        return self.__cake_name

    @property
    def __cake_anon__(self) -> bool:
        return not self.__cake_name

    @property
    def __cake_baked__(self) -> bool:
        return self.__cake_is_baked

    @property
    def __cake_baking_method__(self) -> BakingMethod:
        return self.__cake_baking_method

    @property
    def __cake_recipe__(self) -> R:
        return self.__cake_recipe

    @property
    def __cake_recipe_args__(self) -> tuple:
        return self.__cake_recipe_args

    @property
    def __cake_recipe_kwargs__(self) -> dict:
        return self.__cake_recipe_kwargs

    @property
    def __cake_undefined__(self) -> bool:
        return self.__cake_recipe is UNDEFINED

    @contextmanager
    def __cake_replace__(
        self,
        _cake_recipe: Any,
        *_cake_recipe_args: Any,
        _cake_baking_method: BakingMethod = BakingMethod.BAKE_AUTO,
        **_cake_recipe_kwargs: Any,
    ) -> Iterator[Self]:
        if self.__cake_replaced:
            msg = f"Cannot replace cake '{self}' that's already replaced"
            raise TypeError(msg)

        if self.__cake_is_baked and self.__cake_baking_method != BakingMethod.BAKE_NO_BAKE:
            msg = f"Cannot replace cake '{self}' that's already baked."
            raise TypeError(msg)

        is_replacement: bool = self.__cake_recipe is not UNDEFINED
        orig_recipe_fmt: str = recipe_format(self.__cake_recipe, self.__cake_baking_method)

        self.__cake_replaced = self.__copy__()
        self.__cake_recipe = _cake_recipe  # type: ignore[misc]
        self.__cake_recipe_args = _cake_recipe_args  # type: ignore[misc]
        self.__cake_recipe_kwargs = _cake_recipe_kwargs  # type: ignore[misc]
        self.__cake_baking_method = _cake_baking_method
        self.__cake_is_baked = False
        self.__cake_result = None

        new_recipe_fmt: str = recipe_format(self.__cake_recipe, self.__cake_baking_method)

        if is_replacement:
            logger.debug(f"{self} was replaced: {orig_recipe_fmt} ==> {new_recipe_fmt}")
        try:
            yield self
        finally:
            self.__cake_recipe = self.__cake_replaced.__cake_recipe__  # type: ignore[misc]
            self.__cake_recipe_args = self.__cake_replaced.__cake_recipe_args__  # type: ignore[misc]
            self.__cake_recipe_kwargs = self.__cake_replaced.__cake_recipe_kwargs__  # type: ignore[misc]
            self.__cake_baking_method = self.__cake_replaced.__cake_baking_method__
            self.__cake_replaced = None
            if is_replacement:
                logger.debug(f"{self} was restored: {orig_recipe_fmt} <== {new_recipe_fmt}")

        if self.__cake_is_baked:
            msg = (
                f"'{self}' replacement was rollbacked but cake is still baked. "
                "Unbake the cake first!"
            )
            raise TypeError(msg)

    def __repr__(self) -> str:
        name: str = self.__cake_name or "<anon>"
        return f"Cake '{name}'"

    def __copy__(self) -> Pastry[R]:
        """Copy itself with all ingredients and technologies.

        'Copy cake' just sounds like 'cupcake'. And I like cupcakes ;)
        """
        return Pastry(
            self.__cake_recipe,
            *list(self.__cake_recipe_args),
            _cake_baking_method=self.__cake_baking_method,
            _cake_name=self.__cake_name,
            **dict(self.__cake_recipe_kwargs),
        )

    def __deepcopy__(self, memo: Any) -> Pastry[R]:
        """Deep copy with all ingredients and technologies."""
        return Pastry(
            self.__cake_recipe,
            *deepcopy(self.__cake_recipe_args),
            _cake_baking_method=self.__cake_baking_method,
            _cake_name=self.__cake_name,
            **deepcopy(self.__cake_recipe_kwargs),
        )

    def __call__(self) -> Any:
        assert_baked(self)
        return self.__cake_result

    def __getattr__(self, piece_name: str) -> PieceOfCake:
        """Cut a piece of cake.

        With __CAKE_ATTRIBUTE_ERROR_NAMES__ you can no longer write
        the things like:
        class MyBakery(Bakery):
            ctrl: Any = Cake(Controller)
            controller_func: Any = Cake(some_func, ctrl.func)  # or __wrapped__

        There is the place where AttribureError exception will occured.
        You can make a simple refactoring:
        class MyBakery(Bakery):
            ctrl: Any = Cake(Controller)
            controller_func: Any = Cake(some_func, Cake(getattr, ctrl, "func"))
        """
        if piece_name in self.__CAKE_ATTRIBUTE_ERROR_NAMES__:
            raise AttributeError(piece_name)
        # explicit __getattr__ call to avoid collisions
        return PieceOfCake(self).__getattr__(piece_name)

    def __getitem__(self, piece_name: Any) -> PieceOfCake:
        return PieceOfCake(self).__getitem__(piece_name)

    async def __aenter__(self) -> Any:
        if self.__cake_is_baked:
            return self.__cake_result

        for recipe in flatten(
            [self.__cake_recipe_args, self.__cake_recipe_kwargs, self.__cake_recipe]
        ):
            if is_cake(recipe):
                await recipe.__aenter__()

        recipe = self.__cake_recipe
        if is_cake_or_piece(recipe):
            recipe = recipe()  # type: ignore[operator]

        if not self.__cake_baking_method:
            self.__cake_baking_method = determine_baking_method(recipe)

        self.__cake_result = await bake_recipe(
            recipe,
            recipe_args=self.__cake_recipe_args,
            recipe_kwargs=self.__cake_recipe_kwargs,
            baking_method=self.__cake_baking_method,
            cake_name=str(self),
        )

        logger.debug(f"{self} is baked [{self.__cake_baking_method.name}]")
        self.__cake_is_baked = True
        return self.__cake_result

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Unbake anonymous recipes even if not self.__cake_is_baked.

        Save recipe called value (recipe())
        because it will be impossible to do it after recipe unbaked.
        """
        recipe: Any = self.__cake_recipe

        is_cake_and_baked: bool = is_cake(recipe) and is_baked(recipe)
        is_poc_and_baked: bool = is_piece_of_cake(recipe) and is_baked(recipe.cake)

        if is_cake_and_baked or is_poc_and_baked:
            recipe = recipe()

        _recipe: Any
        for _recipe in flatten(
            [self.__cake_recipe, self.__cake_recipe_args, self.__cake_recipe_kwargs]
        ):
            if is_cake(_recipe) and _recipe.__cake_anon__:
                # unbake anonymous recipes only
                await _recipe.__aexit__(exc_type, exc_value, traceback)

        if not self.__cake_is_baked:
            return

        # All cakes and piece_of_cakes should already be unbaked
        # at this moment
        if not is_cake_or_piece(recipe):
            if self.__cake_baking_method == BakingMethod.BAKE_FROM_CM:
                recipe.__exit__(exc_type, exc_value, traceback)

            elif self.__cake_baking_method == BakingMethod.BAKE_FROM_ACM:
                await recipe.__aexit__(exc_type, exc_value, traceback)

        logger.debug(f"{self} is unbaked")

        self.__cake_is_baked = False


T = TypeVar("T")
P = ParamSpec("P")


def hand_made(cake: T, cake_baking_method: BakingMethod) -> T:
    """Hand made cake."""
    if not is_cake(cake):
        cake = Cake(cake)

    cake._Pastry__cake_baking_method = cake_baking_method  # type: ignore[attr-defined]
    return cake


@overload
def Cake(recipe: Awaitable[T]) -> T: ...


@overload
def Cake(recipe: AbstractAsyncContextManager[T]) -> T: ...


@overload
def Cake(recipe: AbstractContextManager[T]) -> T: ...


@overload
def Cake(
    recipe: Callable[P, Awaitable[T]],
    *recipe_args: P.args,
    **recipe_kwargs: P.kwargs,
) -> T: ...


@overload
def Cake(
    recipe: Callable[P, T],
    *recipe_args: P.args,
    **recipe_kwargs: P.kwargs,
) -> T: ...


@overload
def Cake(recipe: T) -> T: ...


def Cake(  # waiting for issue to close  # noqa: N802
    # https://github.com/python/mypy/issues/11004
    recipe: T,
    *recipe_args: Any,
    **recipe_kwargs: Any,
) -> T:
    return cast(
        "T",
        Pastry(recipe, *recipe_args, **recipe_kwargs),
    )


def __Cake__(*, init: Literal[True] = True) -> Any:  # noqa: N802, N807, ARG001
    return Cake(UNDEFINED)



from __future__ import annotations

__all__ = [
    "anon_cake",
    "assert_baked",
    "cake_baking_method",
    "cake_name",
    "cake_recipe",
    "cake_recipe_args",
    "cake_recipe_kwargs",
    "is_baked",
    "is_cake",
    "is_cake_or_piece",
    "is_piece_of_cake",
    "recipe_format",
]

from typing import (
    TYPE_CHECKING,
    Any,
)

from .types import UNDEFINED, Cakeable, CakeRecipe, FictionalPiece

if TYPE_CHECKING:
    from enum import IntEnum


def is_cake(value: Any) -> bool:
    return isinstance(value, CakeRecipe)


def is_piece_of_cake(value: Any) -> bool:
    return isinstance(value, FictionalPiece)


def is_cake_or_piece(value: Any) -> bool:
    return is_cake(value) or is_piece_of_cake(value)


def cake_name(cake: Cakeable) -> str:
    return cake.__cake_name__


def anon_cake(cake: Cakeable) -> bool:
    return cake.__cake_anon__


def is_baked(cake: Cakeable) -> bool:
    return cake.__cake_baked__


def assert_baked(cake: Cakeable[Any]) -> None:
    if not is_baked(cake):
        msg = f"{cake} is not baked. Just bake it!"
        raise ValueError(msg)


def cake_recipe(cake: Cakeable) -> Any:
    return cake.__cake_recipe__


def cake_recipe_args(cake: Cakeable) -> tuple:
    return cake.__cake_recipe_args__


def cake_recipe_kwargs(cake: Cakeable) -> dict:
    return cake.__cake_recipe_kwargs__


def cake_baking_method(cake: Cakeable) -> IntEnum:
    return cake.__cake_baking_method__


def recipe_format(recipe: Any, method: IntEnum) -> str:
    fmt: str
    if recipe is UNDEFINED:
        fmt = "__cake__"
    elif hasattr(recipe, "__qualname__"):
        fmt = recipe.__qualname__
    elif hasattr(recipe, "__name__"):
        fmt = recipe.__name__
    elif hasattr(recipe, "__class__"):
        fmt = recipe.__class__.__qualname__
    else:
        fmt = str(recipe)

    return f"{fmt}[{method.name}]"



"""Bakery mypy plugin.

Make cakeable all bakery items.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Final

from mypy.plugin import AttributeContext, Plugin
from mypy.types import CallableType, Instance
from mypy.types import Type as MypyType

if TYPE_CHECKING:
    from collections.abc import Callable

    from mypy.options import Options

BAKERY_FULLNAME: Final[str] = "bakery.bakery.Bakery"
CAKEABLE_FULLNAME: Final[str] = "bakery.Cakeable"


def plugin(_: str) -> type[Plugin]:
    """Plugin."""
    return BakeryPlugin


class BakeryPlugin(Plugin):
    """Bakery plugin."""

    def __init__(self, options: Options) -> None:
        super().__init__(options)
        if options.config_file is None:
            return
        # some manipulations on config_file
        # some optimizations here
        self.visible_modules: Final[tuple[str, ...]] = tuple(
            str(path.with_suffix("")).replace("/", ".") for path in Path().rglob("*.py")
        )

    def get_class_attribute_hook(self, _fullname: str) -> Callable | None:
        """Get class attribute hook."""
        # fullname including attribute name
        # e.g. test_app.bakery.MyBakery.some_cake
        # if _fullname.startswith(self.visible_modules):
        return self.class_attribute_hook

    def class_attribute_hook(self, ctx: AttributeContext) -> MypyType:
        """Attribute context."""
        if not (isinstance(ctx.type, CallableType) and isinstance(ctx.type.ret_type, Instance)):
            return ctx.default_attr_type

        for base in ctx.type.ret_type.type.bases:
            if base.type.fullname == BAKERY_FULLNAME:
                break
        else:
            return ctx.default_attr_type

        if ctx.context.name in ("aopen", "aclose", "__aenter__", "__aexit__"):  # type: ignore[attr-defined]
            return ctx.default_attr_type

        smth_inst: Instance = ctx.api.named_type(CAKEABLE_FULLNAME).copy_modified(  # type: ignore[attr-defined]
            args=(ctx.default_attr_type,),
        )
        return smth_inst



from __future__ import annotations

__all__ = [
    "_LOGGER",
    "BakeryLogger",
    "CakeRecipe",
    "DefaultLogger",
]

from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING, Any, Final, Protocol

import bakery

from .types import CakeRecipe

if TYPE_CHECKING:
    from collections.abc import Callable


class BakeryLogger(Protocol):
    """Bakery logger."""

    def debug(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Debug."""

    def info(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Info."""

    def warning(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Warning."""

    def error(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Error."""


class DummyLogger:
    """Dummy logger."""

    def debug(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Debug."""

    def info(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Info."""

    def warning(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Warning."""

    def error(self, message: str, /, *args: Any, **kwargs: Any) -> None:
        """Error."""


class DefaultLogger:
    """Just instead of print function."""

    FUNC_2_LEVEL: Final[dict[str, str]] = {
        "debug": "DEBUG",
        "info": "INFO ",
        "warning": "WARN ",
        "error": "ERROR",
    }

    @staticmethod
    def _log(level: str, message: str) -> None:
        """Log it."""
        print(  # noqa: T201
            f"{datetime.now().isoformat(sep=' ', timespec='milliseconds')} | {level} | {message}"  # noqa: DTZ005
        )

    def __getattr__(self, attr: str) -> Callable[..., Any]:
        """Check and get default logger attribute."""
        if attr not in self.FUNC_2_LEVEL:
            msg = f"Logger has no attribute {attr}. Possible values: {list(self.FUNC_2_LEVEL)}"
            raise AttributeError(
                msg,
            )

        return partial(self._log, self.FUNC_2_LEVEL[attr])


DUMMY_LOGGER: Final[DummyLogger] = DummyLogger()


class LoggerWrapper:
    """Logger wrapper."""

    def __getattr__(self, attr: str) -> Any:
        """Return actual logger method."""
        cur_logger: BakeryLogger = bakery.logger or DUMMY_LOGGER
        return getattr(cur_logger, attr)


_LOGGER: Final[LoggerWrapper] = LoggerWrapper()



from __future__ import annotations

__all__ = [
    "CakeRecipe",
    "Cakeable",
    "FictionalPiece",
    "is_undefined",
]

from inspect import Signature
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Protocol,
    TypeVar,
    final,
)

if TYPE_CHECKING:
    import types
    from enum import IntEnum


class CakeRecipe:
    # for fastapi Depends
    __signature__ = Signature()

    def __call__(self) -> Any:
        raise NotImplementedError


class FictionalPiece:
    # for fastapi Depends
    __signature__ = Signature()

    def __call__(self) -> Any:
        raise NotImplementedError


T_co = TypeVar("T_co", covariant=True)


class PieceProto(Protocol):
    def __getattr__(self, mark: Any) -> PieceProto:
        """some_cake.attribute."""

    def __getitem__(self, mark: Any) -> PieceProto:
        """some_cake["key"]."""

    def __call__(self) -> Any:
        """some_cake()."""


class Cakeable(Protocol[T_co]):
    def __set_name__(self, _: Any, name: str) -> None: ...

    def __call__(self) -> T_co: ...

    def __getattr__(self, piece_name: str) -> PieceProto: ...

    def __getitem__(self, piece_name: Any) -> PieceProto: ...

    @property
    def __cake_name__(self) -> str: ...

    @property
    def __cake_anon__(self) -> bool: ...

    @property
    def __cake_baked__(self) -> bool: ...

    @property
    def __cake_baking_method__(self) -> IntEnum: ...

    @property
    def __cake_recipe__(self) -> Any: ...

    @property
    def __cake_recipe_args__(self) -> tuple: ...

    @property
    def __cake_recipe_kwargs__(self) -> dict: ...

    async def __aenter__(self) -> T_co: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None: ...


@final
class _Undefined: ...


UNDEFINED: Final = _Undefined()


def is_undefined(obj: Any) -> bool:
    return obj is UNDEFINED



from __future__ import annotations

__all__ = [
    "BUILTIN_TYPES",
    "flatten",
    "is_iterable",
    "is_mapping",
    "replace_cakes",
]

import types
from collections.abc import Iterable, Iterator, Mapping
from copy import copy
from typing import (
    Any,
    Final,
    TypeGuard,
)

from .cake_stuff import is_cake_or_piece
from .types import CakeRecipe, FictionalPiece


def is_iterable(
    value: Any,
    ignore_types: tuple[type[Any], ...] = (
        str,
        bytes,
        memoryview,
        types.GeneratorType,
        types.AsyncGeneratorType,
        map,
        range,
        zip,
        CakeRecipe,
        FictionalPiece,
    ),
) -> bool:
    """Check if iterable, ignore generator-like by default."""
    return isinstance(value, Iterable) and not isinstance(value, ignore_types)


def is_mapping(value: Any) -> TypeGuard[Mapping[Any, Any]]:
    """Is mapping check."""
    return isinstance(value, Mapping)


def flatten(
    items: Iterable[Any] | Mapping[Any, Any],
    ignore_types: tuple[type[Any], ...] = (
        str,
        bytes,
        memoryview,
        types.GeneratorType,
        types.AsyncGeneratorType,
        map,
        range,
        zip,
        CakeRecipe,
        FictionalPiece,
    ),
) -> Iterator[Any]:
    """Flatten items, except ignored (generator-like)."""
    items = items.items() if is_mapping(items) else items
    for _item in items:
        if is_iterable(_item, ignore_types) or is_mapping(_item):
            yield from flatten(_item)
        else:
            yield _item


def replace_cakes(obj: Any) -> Any:
    """Replace all objects."""
    res: Any

    if is_mapping(obj):
        res = copy(obj)
        for key, value in obj.items():
            _key = replace_cakes(key)
            _value = replace_cakes(value)
            res[_key] = _value

    elif is_iterable(obj):
        res = type(obj)(replace_cakes(item) for item in obj)

    elif is_cake_or_piece(obj):
        res = obj()

    else:
        res = obj

    return res


BUILTIN_TYPES: Final[tuple[Any, ...]] = (
    bool,
    bytearray,
    bytes,
    classmethod,
    complex,
    dict,
    float,
    frozenset,
    int,
    list,
    map,
    memoryview,
    property,
    range,
    set,
    slice,
    staticmethod,
    str,
    super,
    tuple,
    zip,
    type(None),
    type(Ellipsis),
    types.AsyncGeneratorType,
    types.GeneratorType,
    types.ModuleType,
    types.SimpleNamespace,
)



"""All relative to baking."""

from __future__ import annotations

__all__ = [
    "BakingMethod",
    "bake",
    "bake_recipe",
    "check_baking_method",
    "determine_baking_method",
    "unbake",
]
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from enum import IntEnum, auto
from inspect import isawaitable, iscoroutinefunction
from typing import TYPE_CHECKING, Any, Final, TypeVar

from .stuff import _LOGGER as logger  # noqa: N811
from .stuff import (
    BUILTIN_TYPES,
    is_cake_or_piece,
    is_undefined,
    replace_cakes,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class BakingMethod(IntEnum):
    BAKE_AUTO = 0
    BAKE_FROM_CORO_FUNC = auto()
    BAKE_FROM_AWAITABLE = auto()
    BAKE_FROM_ACM = auto()
    BAKE_FROM_CM = auto()
    BAKE_FROM_BUILTIN = auto()
    BAKE_FROM_CALL = auto()
    BAKE_NO_BAKE = auto()


BAKING_METHODS: Final[dict[BakingMethod, Callable[[Any], bool]]] = {
    BakingMethod.BAKE_FROM_CALL: callable,
    BakingMethod.BAKE_FROM_CM: lambda _recipe: isinstance(_recipe, AbstractContextManager),
    BakingMethod.BAKE_FROM_ACM: lambda _recipe: isinstance(_recipe, AbstractAsyncContextManager),
    BakingMethod.BAKE_FROM_BUILTIN: lambda _recipe: isinstance(_recipe, BUILTIN_TYPES),
    BakingMethod.BAKE_FROM_CORO_FUNC: iscoroutinefunction,
    BakingMethod.BAKE_FROM_AWAITABLE: isawaitable,
    BakingMethod.BAKE_NO_BAKE: lambda _recipe: True,
}


def determine_baking_method(recipe: Any) -> BakingMethod:
    """Determine the first available baking method for recipe."""
    if is_undefined(recipe):
        return BakingMethod.BAKE_NO_BAKE
    if is_cake_or_piece(recipe):
        # Baking method depends on what object
        # really is inside cake/piece_of_cake
        return BakingMethod.BAKE_AUTO

    for method in [
        BakingMethod.BAKE_FROM_CORO_FUNC,
        BakingMethod.BAKE_FROM_AWAITABLE,
        BakingMethod.BAKE_FROM_ACM,
        BakingMethod.BAKE_FROM_CM,
        BakingMethod.BAKE_FROM_BUILTIN,
        BakingMethod.BAKE_FROM_CALL,
    ]:
        if check_baking_method(recipe, method):
            return method

    logger.info(f"Cannot determine baking method for recipe {recipe}")
    return BakingMethod.BAKE_NO_BAKE


def check_baking_method(recipe: Any, method: BakingMethod) -> bool:
    """Check baking method.

    return true if OK.
    """
    try:
        return BAKING_METHODS[method](recipe)
    except KeyError:
        msg = f"Cannot validate unknown baking method '{method}'."
        raise ValueError(msg) from None


async def bake_from_builtin(recipe: Any, _args: Any, _kwargs: Any) -> Any:
    return replace_cakes(recipe)


async def bake_from_call(recipe: Any, args: Any, kwargs: Any) -> Any:
    return recipe(*replace_cakes(args), **replace_cakes(kwargs))


async def bake_from_cm(recipe: Any, _args: Any, _kwargs: Any) -> Any:
    return recipe.__enter__()


async def bake_from_acm(recipe: Any, _args: Any, _kwargs: Any) -> Any:
    return await recipe.__aenter__()


async def bake_from_coro_func(recipe: Any, args: Any, kwargs: Any) -> Any:
    return await recipe(*replace_cakes(args), **replace_cakes(kwargs))


async def bake_from_awaitable(recipe: Any, _args: Any, _kwargs: Any) -> Any:
    return await recipe


async def bake_no_bake(recipe: Any, _args: Any, _kwargs: Any) -> Any:
    return recipe


METHOD_2_HOW_TO_BAKE: Final = {
    BakingMethod.BAKE_FROM_BUILTIN: bake_from_builtin,
    BakingMethod.BAKE_FROM_CALL: bake_from_call,
    BakingMethod.BAKE_FROM_CM: bake_from_cm,
    BakingMethod.BAKE_FROM_ACM: bake_from_acm,
    BakingMethod.BAKE_FROM_CORO_FUNC: bake_from_coro_func,
    BakingMethod.BAKE_FROM_AWAITABLE: bake_from_awaitable,
    BakingMethod.BAKE_NO_BAKE: bake_no_bake,
}


async def bake_recipe(
    recipe: Any,
    *,
    recipe_args: Any,
    recipe_kwargs: Any,
    baking_method: BakingMethod,
    cake_name: str,
) -> Any:
    if baking_method not in METHOD_2_HOW_TO_BAKE:
        msg = f"{cake_name}: Unknown baking method '{baking_method}' for recipe {recipe}"
        raise ValueError(msg)

    return await METHOD_2_HOW_TO_BAKE[baking_method](
        recipe,
        recipe_args,
        recipe_kwargs,
    )


T = TypeVar("T")


async def bake(cake: AbstractAsyncContextManager[T]) -> T:
    return await cake.__aenter__()


async def unbake(
    cake: AbstractAsyncContextManager[T],
    exc_type: type | None = None,
    exc_value: BaseException | None = None,
    traceback: Any | None = None,
) -> None:
    await cake.__aexit__(exc_type, exc_value, traceback)



# isort: skip_file
from typing import Any, TypeVar, Literal
from typing_extensions import dataclass_transform, Self

from .stuff import Cakeable
from .cake import __Cake__

T = TypeVar("T", bound=Bakery)  # noqa: PYI001

def no_init_field(
    *,
    init: Literal[False] = False,
) -> Any: ...

@dataclass_transform(kw_only_default=True, field_specifiers=(no_init_field, __Cake__))
class Bakery:
    __bakery_visitors__: int
    __bakery_items__: dict[str, Cakeable[Any]]
    async def __aenter__(self) -> Self: ...
    async def __aexit__(self, *_args: object) -> None: ...
    @classmethod
    async def aopen(cls) -> Self: ...
    @classmethod
    async def aclose(
        cls,
        *_args: Any,
    ) -> None: ...



"""DI Bakery.

Opened around the clock.
"""

from __future__ import annotations

__all__ = ["Bakery"]

from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import Any, Protocol, TypeVar

from typing_extensions import Self

from .baking import BakingMethod
from .cake import Cake
from .stuff import _LOGGER as logger  # noqa: N811
from .stuff import is_cake

T = TypeVar("T", bound="Bakery")


class Cakeable(Protocol, AbstractAsyncContextManager):
    def __set_name__(self, _: Any, name: str) -> None: ...
    @property
    def __cake_name__(self) -> str: ...
    @property
    def __cake_undefined__(self) -> bool: ...
    def __cake_replace__(
        self,
        _cake_recipe: Any,
        *_cake_recipe_args: Any,
        _cake_baking_method: BakingMethod,
        **_cake_recipe_kwargs: Any,
    ) -> AbstractContextManager[Cakeable]: ...


def replace_cakes(cakes: dict[str, AbstractContextManager]) -> None:
    for replacement in cakes.values():
        replacement.__enter__()


def unreplace_cakes(cakes: dict[str, AbstractContextManager]) -> None:
    for replacement in reversed(cakes.values()):
        replacement.__exit__(None, None, None)

    cakes.clear()


class Bakery:
    """Your bakery."""

    __bakery_visitors__: int
    __bakery_items__: dict[str, Cakeable]
    __bakery_replaced_cakes__: dict[str, AbstractContextManager]

    def __init__(self, **kwargs: Any) -> None:
        cls = type(self)
        if cls.__bakery_visitors__ and kwargs:
            msg = (
                f"{cls.__qualname__} initialized multiple times with keyword arguments. "
                "Such behaviour is discouraging and doesn't make sense. "
                f"Use '{cls.__qualname__}()' instead"
            )
            raise TypeError(msg)

        for item_name, item_value in kwargs.items():
            if item_name not in cls.__bakery_items__:
                msg = f"{cls.__qualname__} got an unexpected keyword argument '{item_name}'"
                raise TypeError(msg)

            if item_name in cls.__bakery_replaced_cakes__:
                msg = (
                    f"{cls.__qualname__} initialized multiple times with keyword argument: "
                    f"'{item_name}'"
                )
                raise TypeError(msg)

            cake_recipe: Any = item_value
            cake_baking_method = BakingMethod.BAKE_NO_BAKE
            recipe_args: tuple = ()
            recipe_kwargs: dict = {}
            if is_cake(item_value):
                cake_recipe = item_value.__cake_recipe__
                recipe_args = item_value.__cake_recipe_args__
                recipe_kwargs = item_value.__cake_recipe_kwargs__
                cake_baking_method = item_value.__cake_baking_method__

            cls.__bakery_replaced_cakes__[item_name] = cls.__bakery_items__[
                item_name
            ].__cake_replace__(
                cake_recipe,
                *recipe_args,
                **recipe_kwargs,
                _cake_baking_method=cake_baking_method,
            )

    async def __aenter__(self) -> Self:
        return await type(self).aopen()

    async def __aexit__(self, *_args: object) -> None:
        return await type(self).aclose()

    def __getattribute__(self, attr: str) -> Any:
        """If getattr(cls, attr) is cake, call it and return."""
        value: Any = super().__getattribute__(attr)
        if is_cake(value):
            return value()
        return value

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize bakery subclass."""
        bakery_items: dict[str, Cakeable] = {}
        # Do filter __dict__, because iterating
        # over __annotations__ forces to annotate
        # every cake
        cake_name: str
        _cake: Cakeable
        for cake_name, _cake in cls.__dict__.items():
            if cake_name.startswith("__") and cake_name.endswith("__"):
                continue
            cake = _cake
            if not is_cake(cake):
                # wrap value into cake
                cake = Cake(cake)
                setattr(cls, cake_name, cake)
                cake.__set_name__(cls, cake_name)
            bakery_items[cake_name] = cake

        cls.__bakery_items__ = bakery_items
        cls.__bakery_visitors__ = 0
        cls.__bakery_replaced_cakes__ = {}

    @classmethod
    async def aopen(cls) -> Self:
        if cls.__bakery_visitors__:
            cls.__bakery_visitors__ += 1
            # no concurrency yet (like aopen/aopen/aopen)
            # anyio lock required (on demand)
            return cls()

        replace_cakes(cls.__bakery_replaced_cakes__)

        missed_args: list[str] = [
            cake.__cake_name__ for cake in cls.__bakery_items__.values() if cake.__cake_undefined__
        ]
        if missed_args:
            msg = (
                f"{cls.__qualname__}.__init__() missing 1 required keyword-only argument: "
                f"'{missed_args[0]}'"
            )
            if len(missed_args) > 1:
                formatted_args: str = (
                    ", ".join(f"'{arg_name}'" for arg_name in missed_args[:-1])
                    + " and "
                    + f"'{missed_args[-1]}'"
                )
                msg = (
                    f"{cls.__qualname__}.__init__() missing {len(missed_args)} required "
                    f"keyword-only arguments: {formatted_args}"
                )

            logger.error(msg)
            unreplace_cakes(cls.__bakery_replaced_cakes__)
            raise TypeError(msg)

        cls.__bakery_visitors__ += 1
        # let's bake all your cakes
        cake: Cakeable | None = None
        try:
            for cake in cls.__bakery_items__.values():
                await cake.__aenter__()
        except (Exception, BaseException) as exc:
            logger.error(f"{cake} cannot be baked: {exc}")
            await cls.aclose()
            raise exc from None

        logger.debug(f"Bakery '{cls.__qualname__}' is opened. Welcome!")
        return cls()

    @classmethod
    async def aclose(
        cls,
        exc_type: type | None = None,
        exc_value: Exception | None = None,
        traceback: Any | None = None,
    ) -> None:
        cls.__bakery_visitors__ -= 1

        if cls.__bakery_visitors__ > 0:
            logger.debug(
                f"Bakery '{cls.__qualname__}' is working till the last visitor "
                f"({cls.__bakery_visitors__} left)!"
            )
            return

        exceptions: list[Exception | BaseException] = []
        cake: Cakeable
        for cake in reversed(  # it's important to unbake in reverse order
            # dict views are reversible since 3.8
            # https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects
            cls.__bakery_items__.values(),
        ):
            try:
                await cake.__aexit__(exc_type, exc_value, traceback)
            except (Exception, BaseException) as exc:  # noqa: PERF203, BLE001
                exceptions.append(exc)

        unreplace_cakes(cls.__bakery_replaced_cakes__)

        logger.debug(f"Bakery '{cls.__qualname__}' is closed. Goodbye!")

        if exceptions:
            # For now raise the first exception occurred.
            # Use trio MultiError (inception) on demand
            # https://github.com/python-trio/trio/blob/v0.21.0/trio/_core/_multierror.py#L154
            # or use exception groups (python 3.11)
            # https://peps.python.org/pep-0654/
            raise exceptions[0]



"""Test bakery."""

from __future__ import annotations

import warnings
from inspect import Parameter, Signature
from typing import TYPE_CHECKING, Any

import pytest

from . import Bakery as _Bakery
from . import (
    Cake,
    Cakeable,
    cake_baking_method,
    cake_recipe,
    cake_recipe_args,
    cake_recipe_kwargs,
    is_cake,
    unbake,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from pytest_mock import MockerFixture

Bakery = type[_Bakery]


def fixture_factory(mocker_name: str) -> Any:
    """Create bakery mock fixtures by mocker name."""

    async def mocker(**kwargs: Any) -> AsyncIterator[BakeryMock]:
        if not all((len(kwargs) == 1, next(iter(kwargs.keys())).endswith("mocker"))):
            msg: str = "Please, pass mocker fixture."
            raise ValueError(msg)

        mock: BakeryMock = BakeryMock(next(iter(kwargs.values())))
        yield mock
        await mock.stopall()

    mocker.__signature__ = Signature(  # type: ignore[attr-defined]
        [
            Parameter(
                mocker_name,
                kind=Parameter.KEYWORD_ONLY,
            ),
        ],
    )
    return mocker


bakery_mock = pytest.fixture()(fixture_factory("mocker"))
class_bakery_mock = pytest.fixture(scope="class")(fixture_factory("class_mocker"))
module_bakery_mock = pytest.fixture(scope="module")(fixture_factory("module_mocker"))
package_bakery_mock = pytest.fixture(scope="package")(fixture_factory("package_mocker"))
session_bakery_mock = pytest.fixture(scope="session")(fixture_factory("session_mocker"))


class BakeryMock:
    """Toy bakery."""

    def __init__(self, mocker: MockerFixture) -> None:
        self._mocker_: MockerFixture = mocker
        self._cake_mocks_: dict[str, Cakeable[Any]] = {}
        self._bakery_: Bakery | None = None
        self._children_: list[BakeryMock] = []

    def __copy__(self) -> BakeryMock:
        """Copy bakery mock."""
        mock_copy: BakeryMock = BakeryMock(self._mocker_)
        self._children_.append(mock_copy)
        return mock_copy

    async def _patch_cake(
        self,
        *,
        bakery: Bakery,
        cake: Cakeable[Any],
        new_recipe: Cakeable[Any],
    ) -> None:
        """Patch cake recipe."""
        if not all((is_cake(cake), is_cake(new_recipe), issubclass(bakery, _Bakery))):
            raise ValueError

        new_cake: Cakeable[Any] = new_recipe
        self._mocker_.patch.multiple(
            cake,
            _Pastry__cake_recipe=cake_recipe(new_cake),
            _Pastry__cake_recipe_args=cake_recipe_args(new_cake),
            _Pastry__cake_recipe_kwargs=cake_recipe_kwargs(new_cake),
            _Pastry__cake_baking_method=cake_baking_method(new_cake),
            _Pastry__cake_result=await new_cake.__aenter__(),
        )

    async def _patch(self, bakery: Bakery) -> None:
        """Start bakery `bakery` patching."""
        if not issubclass(bakery, _Bakery):
            msg: str = f"{bakery} is not a Bakery."
            raise TypeError(msg)

        cake_name: str
        new_recipe: Any

        for cake_name, new_recipe in self._cake_mocks_.items():
            # patch bakery itself
            await self._patch_cake(
                bakery=bakery,
                cake=getattr(bakery, cake_name),
                new_recipe=new_recipe,
            )

        await bakery.aopen()
        self._bakery_ = bakery

    async def patch(self, bakery: Bakery) -> None:
        """Check if bakery set and start patching."""
        if self._bakery_ is not None:
            msg: str = f"Old bakery {self._bakery_} is still set. Close it first."
            raise ValueError(msg)
        return await self._patch(bakery)

    def __call__(self, bakery: Bakery) -> BakeryMock:
        if self._bakery_ is not None:
            msg: str = f"Old bakery {self._bakery_} is still set. Close it first."
            raise ValueError(msg)
        self._bakery_ = bakery
        return self

    async def __aenter__(self) -> None:
        if self._bakery_ is None:
            msg: str = "You should set bakery first."
            raise ValueError(msg)
        return await self._patch(self._bakery_)

    async def __aexit__(self, *_args: object) -> None:
        return await self.reset()

    async def reset(self) -> None:
        """Stop patching."""
        if self._bakery_:
            await self._bakery_.aclose()
            self._bakery_ = None

        self._mocker_.stopall()

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in (
            "_mocker_",
            "_cake_mocks_",
            "_bakery_",
            "_children_",
        ):
            return super().__setattr__(attr, value)

        if not is_cake(value):
            warnings.warn(
                "Direct value assignment to bakery_mock is deprecated and will be removed soon. "
                "Please, use bakery_mock.attr = Cake(value) notation.",
                stacklevel=2,
            )
            value = Cake(value)
        self._cake_mocks_[attr] = value
        return None

    async def stopall(self) -> None:
        """Reset mocker and clear piece mocks."""
        await self.reset()
        for cake in self._cake_mocks_.values():
            await unbake(cake)

        for child in self._children_:
            await child.stopall()

        self._cake_mocks_.clear()
        self._children_.clear()
